@@ -23,8 +23,9 @@ from docspec_python import format_arglist
 from nr.databind.core import Field, Struct, Validator
 from nr.interface import implements, override
 from pydoc_markdown.interfaces import Context, Renderer, Resolver, SourceLinker
-from typing import Iterable, List, Optional, TextIO
+from typing import cast, Iterable, List, Optional, TextIO
 import docspec
+import html
 import io
 import sys
 
@@ -39,9 +40,6 @@ class MarkdownRenderer(Struct):
   ### Options
   """
 
-  #: Can be used to explicitly specify a file object to render to.
-  fp = Field(object, default=None, hidden=True)
-
   #: The name of the file to render to. If no file is specified, it will
   #: render to stdout.
   filename = Field(str, default=None)
@@ -147,6 +145,9 @@ class MarkdownRenderer(Struct):
   #: Render module headers. This is enabled by default.
   render_module_header = Field(bool, default=True)
 
+  #: Custom template for module header.
+  render_module_header_template = Field(str, default='')
+
   #: Render docstrings as blockquotes. This is disabled by default.
   docstrings_as_blockquote = Field(bool, default=False)
 
@@ -177,6 +178,9 @@ class MarkdownRenderer(Struct):
   #: file if a #source_linker is configured. The default is `[[view_source]]({url})`.
   source_format = Field(str, default='[[view_source]]({url})')
 
+  #: Escape html in docstring. Default to False.
+  escape_html_in_docstring = Field(bool, default=False)
+
   _reverse_map = Field(docspec.ReverseMap, default=None, hidden=True)
   def _get_parent(self, obj: docspec.ApiObject) -> Optional[docspec.ApiObject]:
     return self._reverse_map.get_parent(obj)
@@ -201,6 +205,10 @@ class MarkdownRenderer(Struct):
       self._render_toc(fp, level, child)
 
   def _render_header(self, fp, level, obj):
+    if self.render_module_header_template and isinstance(obj, docspec.Module):
+      fp.write(self.render_module_header_template.format(module_name=obj.name))
+      return
+
     object_id = self._generate_object_id(obj)
     if self.use_fixed_header_levels:
       # Read the header level based on the API object type. The default levels defined
@@ -223,7 +231,7 @@ class MarkdownRenderer(Struct):
       yield '@{}{}\n'.format(dec.name, dec.args or '')
 
   def _format_function_signature(self, func: docspec.Function, override_name: str = None, add_method_bar: bool = True) -> str:
-    parts = []
+    parts: List[str] = []
     if self.signature_with_decorators:
       parts += self._format_decorations(func.decorations)
     if self.signature_python_help_style and not self._is_method(func):
@@ -232,7 +240,9 @@ class MarkdownRenderer(Struct):
     if self.signature_with_def:
       parts.append('def ')
     if self.signature_class_prefix and self._is_method(func):
-      parts.append(self._get_parent(func).name + '.')
+      parent = self._get_parent(func)
+      assert parent, func
+      parts.append(parent.name + '.')
     parts.append((override_name or func.name))
     parts.append('(' + self._format_arglist(func) + ')')
     if func.return_type:
@@ -286,7 +296,8 @@ class MarkdownRenderer(Struct):
     if source_string and self.source_position == 'after signature':
       fp.write(source_string + '\n\n')
     if obj.docstring:
-      lines = obj.docstring.split('\n')
+      docstring = html.escape(obj.docstring) if self.escape_html_in_docstring else obj.docstring
+      lines = docstring.split('\n')
       if self.docstrings_as_blockquote:
         lines = ['> ' + x for x in lines]
       fp.write('\n'.join(lines))
@@ -298,18 +309,6 @@ class MarkdownRenderer(Struct):
     for member in getattr(obj, 'members', []):
       self._render_recursive(fp, level, member)
 
-  def _render_modules(self, modules: List[docspec.Module], fp: TextIO):
-    self._reverse_map = docspec.ReverseMap(modules)
-
-    if self.render_toc:
-      if self.render_toc_title:
-        fp.write('# {}\n\n'.format(self.render_toc_title))
-      for m in modules:
-        self._render_toc(fp, 0, m)
-      fp.write('\n')
-    for m in modules:
-      self._render_recursive(fp, 1, m)
-
   def _get_title(self, obj):
     title = obj.name
     if (self.add_method_class_prefix and self._is_method(obj)) or \
@@ -350,9 +349,21 @@ class MarkdownRenderer(Struct):
 
   def render_to_string(self, modules: List[docspec.Module]) -> str:
     fp = io.StringIO()
-    self._render_modules(modules, fp)
+    self.render_to_stream(modules, fp)
     return fp.getvalue()
 
+  def render_to_stream(self, modules: List[docspec.Module], stream: TextIO):
+    self._reverse_map = docspec.ReverseMap(modules)
+
+    if self.render_toc:
+      if self.render_toc_title:
+        stream.write('# {}\n\n'.format(self.render_toc_title))
+      for m in modules:
+        self._render_toc(stream, 0, m)
+      stream.write('\n')
+    for m in modules:
+      self._render_recursive(stream, 1, m)
+
   # Renderer
 
   @override
@@ -390,10 +401,10 @@ class MarkdownRenderer(Struct):
   @override
   def render(self, modules: List[docspec.Module]) -> None:
     if self.filename is None:
-      self._render_modules(modules, self.fp or sys.stdout)
+      self.render_to_stream(modules, sys.stdout)
     else:
       with io.open(self.filename, 'w', encoding=self.encoding) as fp:
-        self._render_modules(modules, fp)
+        self.render_to_stream(modules, cast(TextIO, fp))
 
   # PluginBase
 
